from ... import _hub as inj
pygame = inj.pygame

# TODO model refactoring: use matrices as building blocks,
# it will save some effort (Cell, CellLayer etc.)

# decompress ; b64decode ; struct... Sont tous 3
# employés dans Layer.fromxml
# TODO check if its katasdk -compatible?
from zlib import decompress
from base64 import b64decode
import struct

# ElemenTree est employé en 2 endroits: dans Tileset.fromxml & Tilemap.load
# this works in local ctx:
# from xml.etree import ElementTree
from xml.etree import ElementTree


def load_tmx(filename):
    return TileMap.load(filename)


class Cell:
    """
    Layers are made of Cells (or empty space).
    Cells have some basic properties:

    x, y - the cell's index in the layer
    px, py - the cell's pixel position
    left, right, top, bottom - the cell's pixel boundaries

    Additionally the cell may have other properties which are accessed using
    standard dictionary methods:

    cell['property name']

    You may assign a new value for a property to or even delete an existing
    property from the cell - this will not affect the Tile or any other Cells
    using the Cell's Tile.
    """

    def __init__(self, x, y, px, py, tile):
        self.x, self.y = x, y
        self.px, self.py = px, py
        self.tile = tile
        self.topleft = (px, py)
        self.left = px
        self.right = px + tile.tile_width
        self.top = py
        self.bottom = py + tile.tile_height
        self.center = (px + tile.tile_width // 2, py + tile.tile_height // 2)
        self._added_properties = {}
        self._deleted_properties = set()

    def __repr__(self):
        return '<Cell %s,%s %d>' % (self.px, self.py, self.tile.gid)

    def __contains__(self, key):
        if key in self._deleted_properties:
            return False
        return key in self._added_properties or key in self.tile.properties

    def __getitem__(self, key):
        if key in self._deleted_properties:
            raise KeyError(key)
        if key in self._added_properties:
            return self._added_properties[key]
        if key in self.tile.properties:
            return self.tile.properties[key]
        raise KeyError(key)

    def __setitem__(self, key, value):
        self._added_properties[key] = value

    def __delitem__(self, key):
        self._deleted_properties.add(key)

    def intersects(self, other):
        """
        Determine whether this Cell intersects with the other rect (which has
        .x, .y, .width and .height attributes.)
        """
        if self.px + self.tile.tile_width < other.x:
            return False
        if other.x + other.width - 1 < self.px:
            return False
        if self.py + self.tile.tile_height < other.y:
            return False
        if other.y + other.height - 1 < self.py:
            return False
        return True


class LayerIterator(object):
    """
    Iterates over all the cells in a layer in column,row order.
    """

    def __init__(self, layer):
        self.layer = layer
        self.i, self.j = 0, 0

    def __next__(self):
        if self.i == self.layer.width - 1:
            self.j += 1
            self.i = 0
        if self.j == self.layer.height - 1:
            raise StopIteration()
        value = self.layer[self.i, self.j]
        self.i += 1
        return value


# TODO faut sortir draw
class ObjectTmx:
    """An object in a TMX object layer.
    name: The name of the object. An arbitrary string.
    type: The type of the object. An arbitrary string.
    x: The x coordinate of the object in pixels.
    y: The y coordinate of the object in pixels.
    width: The width of the object in pixels (defaults to 0).
    height: The height of the object in pixels (defaults to 0).
    gid: An reference to a tile (optional).
    visible: Whether the object is shown (1) or hidden (0). Defaults to 1.
    """

    def __init__(self, type, x, y, width=0, height=0, name=None,
                 gid=None, tile=None, visible=1):
        self.type = type
        self.px = x
        self.left = x
        if tile:
            y -= tile.tile_height
            width = tile.tile_width
            height = tile.tile_height
        self.py = y
        self.top = y
        self.width = width
        self.right = x + width
        self.height = height
        self.bottom = y + height
        self.name = name
        self.gid = gid
        self.tile = tile
        self.visible = visible
        self.properties = {}

        self._added_properties = {}
        self._deleted_properties = set()

    def __repr__(self):
        if self.tile:
            return '<Object %s,%s %s,%s tile=%d>' % (self.px, self.py, self.width, self.height, self.gid)
        else:
            return '<Object %s,%s %s,%s>' % (self.px, self.py, self.width, self.height)

    def __contains__(self, key):
        if key in self._deleted_properties:
            return False
        if key in self._added_properties:
            return True
        if key in self.properties:
            return True
        return self.tile and key in self.tile.properties

    def __getitem__(self, key):
        if key in self._deleted_properties:
            raise KeyError(key)
        if key in self._added_properties:
            return self._added_properties[key]
        if key in self.properties:
            return self.properties[key]
        if self.tile and key in self.tile.properties:
            return self.tile.properties[key]
        raise KeyError(key)

    def __setitem__(self, key, value):
        self._added_properties[key] = value

    def __delitem__(self, key):
        self._deleted_properties.add(key)

    def draw(self, surface, view_x, view_y):
        if self.visible:
            x, y = (self.px - view_x, self.py - view_y)
            if self.tile:
                surface.blit(self.tile.surface, (x, y))
            else:
                r = pygame.Rect((x, y), (self.width, self.height))
                pygame.draw.rect(surface, (255, 100, 100), r, 2)

    @classmethod
    def fromxml(cls, tag, map):
        if 'gid' in tag.attrib:
            gid = int(tag.attrib['gid'])
            tile = map.tilesets[gid]
            w = tile.tile_width
            h = tile.tile_height
        else:
            gid = None
            tile = None
            w = round(float(tag.attrib['width']))
            h = round(float(tag.attrib['height']))

        x_attr, y_attr = round(float(tag.attrib['x'])), round(float(tag.attrib['y']))
        o = cls(
            tag.attrib.get('type', 'rect'),
            x_attr, y_attr, w, h, tag.attrib.get('name'), gid, tile,
            int(tag.attrib.get('visible', 1))
        )

        props = tag.find('properties')
        if props is None:
            return o

        for c in props.findall('property'):
            # store additional properties.
            name = c.attrib['name']
            value = c.attrib['value']

            # TODO hax
            if value.isdigit():
                value = int(value)
            o.properties[name] = value
        return o

    def intersects(self, x1, y1, x2, y2):
        return not any((x2 < self.px, y2 < self.py, x1 > self.px + self.width, y1 > self.py + self.height))


# TODO la aussi, deux methodes de dessin a evacuer!
class ObjectLayer:
    """
    A layer composed of basic primitive shapes.
    Actually encompasses a TMX <objectgroup> but even the TMX documentation
    refers to them as object layers, so I will.

    ObjectLayers have some basic properties:

        position - ignored (cannot be edited in the current Tiled editor)
        name - the name of the object group.
        color - the color used to display the objects in this group.
        opacity - the opacity of the layer as a value from 0 to 1.
        visible - whether the layer is shown (1) or hidden (0).
        objects - the objects in this Layer (Object instances)
    """

    def __init__(self, name, color, objects, opacity=1,
                 visible=1, position=(0, 0)):
        self.name = name
        self.color = color
        self.objects = objects
        self.opacity = opacity
        self.visible = visible
        self.position = position
        self.properties = {}

    def __repr__(self):
        return '<ObjectLayer "%s" at 0x%x>' % (self.name, id(self))

    @classmethod
    def fromxml(cls, tag, map):
        layer = cls(tag.attrib['name'], tag.attrib.get('color'), [],
                    float(tag.attrib.get('opacity', 1)),
                    int(tag.attrib.get('visible', 1)))
        for object in tag.findall('object'):
            layer.objects.append(ObjectTmx.fromxml(object, map))
        for c in tag.findall('property'):
            # store additional properties.
            name = c.attrib['name']
            value = c.attrib['value']

            # TODO hax
            if value.isdigit():
                value = int(value)
            layer.properties[name] = value
        return layer

    def update(self, dt, *args):
        pass

    def set_view(self, x, y, w, h, viewport_ox=0, viewport_oy=0):
        self.view_x, self.view_y = x, y
        self.view_w, self.view_h = w, h
        x -= viewport_ox
        y -= viewport_oy
        self.position = (x, y)

    def draw(self, surface):
        """
        Draw this layer, limited to the current viewport, to the Surface.
        """
        if not self.visible:
            return
        ox, oy = self.position
        w, h = self.view_w, self.view_h
        for myobj in self.objects:
            myobj.draw(surface, self.view_x, self.view_y)

    def find(self, *properties):
        """
        Find all cells with the given properties set.
        """
        r = []
        for propname in properties:
            for myobj in self.objects:
                if myobj and propname in myobj or propname in self.properties:
                    r.append(myobj)
        return r

    def match(self, **properties):
        """
        Find all objects with the given properties set to the given values.
        """
        r = []
        for propname in properties:
            for myobj in self.objects:
                if propname in myobj:
                    val = myobj[propname]
                elif propname in self.properties:
                    val = self.properties[propname]
                else:
                    continue
                if properties[propname] == val:
                    r.append(myobj)
        return r

    def collide(self, rect, propname):
        """
        Find all objects the rect is touching that have the indicated
        property name set.
        """
        r = []
        for myobj in self.get_in_region(rect.left, rect.top, rect.right, rect.bottom):
            if propname in myobj or propname in self.properties:
                r.append(myobj)
        return r

    def get_in_region(self, x1, y1, x2, y2):
        """
        Return objects that are within the map-space
        pixel bounds specified by the bottom-left (x1, y1) and top-right
        (x2, y2) corners.

        Return a list of Object instances.
        """
        return [obj for obj in self.objects if obj.intersects(x1, y1, x2, y2)]

    def get_at(self, x, y):
        """
        Return the first object found at the nominated (x, y) coordinate.
        Return an Object instance or None.
        """
        for myobj in self.objects:
            if myobj.contains(x, y):
                return myobj


# TODO y a deux methodes a evacuer c'est set_view(...) et draw(...)
class Layer:
    """
    A 2d grid of Cells.

    Layers have some basic properties:

        width, height - the dimensions of the Layer in cells
        tile_width, tile_height - the dimensions of each cell
        px_width, px_height - the dimensions of the Layer in pixels
        tilesets - the tilesets used in this Layer (a Tilesets instance)
        properties - any properties set for this Layer
        cells - a dict of all the Cell instances for this Layer, keyed off
                (x, y) index.

    Additionally you may look up a cell using direct item access:

       layer[x, y] is layer.cells[x, y]

    Note that empty cells will be set to None instead of a Cell instance.
    """

    def __init__(self, name, visible, map):
        self.name = name
        self.visible = visible
        self.position = (0, 0)
        # TODO get from TMX?
        self.px_width = map.px_width
        self.px_height = map.px_height
        self.tile_width = map.tile_width
        self.tile_height = map.tile_height
        self.width = map.width
        self.height = map.height
        self.tilesets = map.tilesets
        self.group = pygame.sprite.Group()
        self.properties = {}
        self.cells = {}

    def __repr__(self):
        return '<Layer "%s" at 0x%x>' % (self.name, id(self))

    def __getitem__(self, pos):
        return self.cells.get(pos)

    def __setitem__(self, pos, tile):
        x, y = pos
        px = x * self.tile_width
        py = y * self.tile_width
        self.cells[pos] = Cell(x, y, px, py, tile)

    def __iter__(self):
        return LayerIterator(self)

    @classmethod
    def fromxml(cls, tag, givenmap):
        layer = cls(tag.attrib['name'], int(tag.attrib.get('visible', 1)), givenmap)

        data = tag.find('data')
        if data is None:
            raise ValueError('layer %s does not contain <data>' % layer.name)

        data = data.text.strip()
        data = data.encode()  # Convert to bytes
        # Decode from base 64 and decompress via zlib
        data = decompress(b64decode(data))
        data = struct.unpack('<%di' % (len(data) / 4,), data)
        assert len(data) == layer.width * layer.height
        for idx, gid in enumerate(data):
            if gid >= 1:  # otherwise its not set
                tile = givenmap.tilesets[gid]
                x = idx % layer.width
                y = idx // layer.width
                layer.cells[x, y] = Cell(x, y, x * givenmap.tile_width, y * givenmap.tile_height, tile)

        return layer

    def update(self, dt, *args):
        pass

    def set_view(self, x, y, w, h, viewport_ox=0, viewport_oy=0):
        self.view_x, self.view_y = x, y
        self.view_w, self.view_h = w, h
        x -= viewport_ox
        y -= viewport_oy
        self.position = (x, y)

    def draw(self, surface):
        """
        Draw this layer, limited to the current viewport, to the Surface.
        """
        ox, oy = self.position
        w, h = self.view_w, self.view_h
        for x in range(ox, ox + w + self.tile_width, self.tile_width):
            i = x // self.tile_width
            for y in range(oy, oy + h + self.tile_height, self.tile_height):
                j = y // self.tile_height
                if (i, j) not in self.cells:
                    continue
                cell = self.cells[i, j]
                surface.blit(cell.tile.surface, (cell.px - ox, cell.py - oy))

    def find(self, *properties):
        """
        Find all cells with the given properties set.
        """
        r = []
        for propname in properties:
            for cell in list(self.cells.values()):
                if cell and propname in cell:
                    r.append(cell)
        return r

    def match(self, **properties):
        """
        Find all cells with the given properties set to the given values.
        """
        r = []
        for propname in properties:
            for cell in list(self.cells.values()):
                if propname not in cell:
                    continue
                if properties[propname] == cell[propname]:
                    r.append(cell)
        return r

    def collide(self, rect, propname):
        """
        Find all cells the rect is touching that have the indicated property
        name set.
        """
        r = []
        for cell in self.get_in_region(rect.left, rect.top, rect.right,
                                       rect.bottom):
            if not cell.intersects(rect):
                continue
            if propname in cell:
                r.append(cell)
        return r

    def get_in_region(self, x1, y1, x2, y2):
        """
        Return cells (in [column][row]) that are within the map-space
        pixel bounds specified by the bottom-left (x1, y1) and top-right
        (x2, y2) corners.
        Return a list of Cell instances.
        """
        i1 = max(0, x1 // self.tile_width)
        j1 = max(0, y1 // self.tile_height)
        i2 = min(self.width, x2 // self.tile_width + 1)
        j2 = min(self.height, y2 // self.tile_height + 1)
        return [self.cells[i, j]
                for i in range(int(i1), int(i2))
                for j in range(int(j1), int(j2))
                if (i, j) in self.cells]

    def get_at(self, x, y):
        """
        Return the cell at the nominated (x, y) coordinate.
        Return a Cell instance or None.
        """
        i = x // self.tile_width
        j = y // self.tile_height
        return self.cells.get((i, j))

    def neighbors(self, index):
        """
        Return the indexes of the valid (ie. within the map) cardinal (ie.
        North, South, East, West) neighbors of the nominated cell index.
        Returns a list of 2-tuple indexes.
        """
        i, j = index
        n = []
        if i < self.width - 1:
            n.append((i + 1, j))
        if i > 0:
            n.append((i - 1, j))
        if j < self.height - 1:
            n.append((i, j + 1))
        if j > 0:
            n.append((i, j - 1))
        return n


class Tile:

    def __init__(self, gid, surface, tileset):
        self.gid = gid
        self.surface = surface
        self.tile_width = tileset.tile_width
        self.tile_height = tileset.tile_height
        self.properties = {}

    @classmethod
    def fromSurface(cls, surface):
        """
        Create a new Tile object straight from a pygame Surface.
        Its tile_width and tile_height will be set using the Surface dimensions.
        Its gid will be 0.
        """
        class ts:
            tile_width, tile_height = surface.get_size()

        return cls(0, surface, ts)

    def loadxml(self, tag):
        props = tag.find('properties')
        if props is None:
            return
        for c in props.findall('property'):
            # store additional properties.
            name = c.attrib['name']
            value = c.attrib['value']

            # TODO hax
            if value.isdigit():
                value = int(value)
            self.properties[name] = value

    def __repr__(self):
        return '<Tile %d>' % self.gid


class Tileset:
    """
    (mm si l'on charge des img pygame ca reste du modele car on fetch les donnees depuis
    une telle instance...)
    """

    def __init__(self, name, tile_width, tile_height, firstgid):
        self.name = name
        self.tile_width = tile_width
        self.tile_height = tile_height
        self.firstgid = firstgid
        self.tiles = []
        self.properties = {}

    def add_image(self, file):
        image = pygame.image.load(file).convert_alpha()
        if not image:
            print("Error creating new Tileset: file %s not found" % file)
            raise FileNotFoundError

        ident = self.firstgid
        tile_w, tile_h = self.tile_width, self.tile_height
        for line in range(image.get_height() // tile_h):
            for column in range(image.get_width() // tile_w):
                pos = pygame.Rect(column * tile_w, line * tile_h, tile_w, tile_h)
                self.tiles.append(Tile(ident, image.subsurface(pos), self))
                ident += 1

    def get_tile(self, gid):
        return self.tiles[gid - self.firstgid]

    @classmethod
    def fromxml(cls, tag, firstgid=None, hacksource=None, hacktileset=None):
        print('fromxml ')
        if 'source' in tag.attrib:
            firstgid = int(tag.attrib['firstgid'])
            if hacksource:
                srcc = hacksource
            else:
                srcc = tag.attrib['source']

            with open(srcc) as f:
                print('opened ', srcc)
                tileset = ElementTree.fromstring(f.read())

            return cls.fromxml(tileset, firstgid, hacktileset=hacktileset)

        name = tag.attrib['name']
        if firstgid is None:
            firstgid = int(tag.attrib['firstgid'])
        tile_width = int(tag.attrib['tilewidth'])
        tile_height = int(tag.attrib['tileheight'])

        tileset = cls(name, tile_width, tile_height, firstgid)

        for c in tag:  # .getchildren():
            if c.tag == "image":
                # create a tileset
                arg_sheet = c.attrib['source'] if (hacktileset is None) else hacktileset
                tileset.add_image(arg_sheet)
            elif c.tag == 'tile':
                gid = tileset.firstgid + int(c.attrib['id'])
                tileset.get_tile(gid).loadxml(c)
        return tileset


class Tilesets(dict):
    def add(self, tileset):
        for idx, tile in enumerate(tileset.tiles):
            idx += tileset.firstgid
            self[idx] = tile


class Layers(list):
    def __init__(self):
        self.by_name = {}

    def add_named(self, layer, name):
        self.append(layer)
        self.by_name[name] = layer

    def __getitem__(self, item):
        if isinstance(item, int):
            return self[item]
        return self.by_name[item]


class TileMap:
    """
    A TileMap is a collection of Layers which contain gridded maps or sprites
    TileMaps are loaded from TMX files which sets the .layers and .tilesets properties

    After loading, additional SpriteLayers may be added.

    A TileMap's rendering is restricted by a viewport which is defined by the
    size passed in at construction time and the focus set by set_focus() or
    force_focus().

    TileMaps have a number of properties:

        width, height - the dimensions of the tilemap in cells

        tile_width, tile_height - the dimensions of the cells in the map
        px_width, px_height - the dimensions of the tilemap in pixels
        properties - any properties set on the tilemap in the TMX file
        layers - all layers of this tilemap as a Layers instance
        tilesets - all tilesets of this tilemap as a Tilesets instance

    """

    def __init__(self):
        self.px_width = 0
        self.px_height = 0
        self.tile_width = 0
        self.tile_height = 0
        self.width = 0
        self.height = 0
        self.properties = {}
        self.layers = Layers()
        self.tilesets = Tilesets()

    @property
    def pix_width(self):
        return self.px_width

    @property
    def pix_height(self):
        return self.px_height

    def update(self, dt, *args):
        for layer in self.layers:
            layer.update(dt, *args)

    @classmethod
    def load(cls, filename, hack_tsxfile=None, hack_ts=None):
        with open(filename) as f:
            tminfo_tree = ElementTree.fromstring(f.read())

        # get most general map informations and create a surface
        tilemap = TileMap()

        tilemap.width = int(tminfo_tree.attrib['width'])
        tilemap.height = int(tminfo_tree.attrib['height'])
        tilemap.tile_width = int(tminfo_tree.attrib['tilewidth'])
        tilemap.tile_height = int(tminfo_tree.attrib['tileheight'])
        tilemap.px_width = tilemap.width * tilemap.tile_width
        tilemap.px_height = tilemap.height * tilemap.tile_height

        for tag in tminfo_tree.findall('tileset'):
            tilemap.tilesets.add(
                Tileset.fromxml(tag, hacksource=hack_tsxfile, hacktileset=hack_ts)  # hacks work only if no more than 1 ts
            )
            print('tilesets added')

        for tag in tminfo_tree.findall('layer'):
            layer = Layer.fromxml(tag, tilemap)
            tilemap.layers.add_named(layer, layer.name)

        for tag in tminfo_tree.findall('objectgroup'):
            layer = ObjectLayer.fromxml(tag, tilemap)
            tilemap.layers.add_named(layer, layer.name)

        # We wanna persist info about the (optional) Image layer found in .tmx
        # TODO add to kengi a class that modelize ImageLayer?
        tagraw_ilayer_li = tminfo_tree.findall('imagelayer')
        if len(tagraw_ilayer_li) == 1:
            tagraw_ilayer = tagraw_ilayer_li.pop()
            subnode = tagraw_ilayer.find('image')
            tilemap.background = {  # persist info in tilemap_obj.background
                'img_path': subnode.attrib['source'],
                'offsetx': round(float(tagraw_ilayer.attrib['offsetx'])),
                'offsety': round(float(tagraw_ilayer.attrib['offsety'])),
                'repeatx': int(tagraw_ilayer.attrib['repeatx'])  # 0 or 1
            }
        elif len(tagraw_ilayer_li) > 1:
            raise NotImplementedError
        else:
            tilemap.background = None

        return tilemap



import random
from ... import _hub as inj
# import pygame


pygame = inj.pygame


class Label(pygame.sprite.Sprite):

    def __init__(self, txt, color=None):
        self._font = pygame.font.Font(None, 35)

        super().__init__()

        self._txt = txt

        self._color = None
        if color is None:
            color = (0, 0, 0)
        self.color = color

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self.image = self._font.render(self._txt, False, value)
        self.rect = self.image.get_rect()


if __name__ == '__main__':
    pygame.init()
    scr = pygame.display.set_mode((400, 300))
    gameover = False
    cl = pygame.time.Clock()
    central_obj = Label('salut', 'steelblue')

    while not gameover:
        for ev in pygame.event.get():
            if ev.type == pygame.QUIT:
                gameover = True
            elif ev.type == pygame.KEYDOWN:
                central_obj.color = random.choice(('orange', 'pink', 'yellow'))

            scr.fill((87, 77, 15))
            scr.blit(central_obj.image, central_obj.rect.topleft)
            pygame.display.update()
            cl.tick(60)
    print('test over.')
    pygame.quit()



import random
import math
from .. import struct


# ----------------------------
#  random map generation tool
# ----------------------------
SYM_UP, SYM_DOWN, SYM_LEFT, SYM_RIGHT = range(4)

DIRS = [
    SYM_UP,
    SYM_DOWN,
    SYM_LEFT,
    SYM_RIGHT]

COORD_OFFSET = {
    SYM_UP: [0, -1],
    SYM_DOWN: [0, +1],
    SYM_LEFT: [-1, 0],
    SYM_RIGHT: [+1, 0]}

WINDING_FACTOR = 0.4
SEUIL_CASES = 44


def dist_manhattan(c1, c2):
    return abs(c2[0] - c1[0]) + abs(c2[1] - c1[1])


def cell_in_range(cell, size):
    if cell[0] < 0 or cell[0] >= size[0]:
        return False
    if cell[1] < 0 or cell[1] >= size[1]:
        return False
    return True


def cell_voisinnes(c, size):
    i, j = c
    res = list()
    for d in DIRS:
        voisin_direct = (i + COORD_OFFSET[d][0], j + COORD_OFFSET[d][1])
        if cell_in_range(voisin_direct, size):
            res.append(voisin_direct)
    return res


class RandomMaze:
    """
    le but de cette classe est de générer un labyrinthe aléatoire :
    représenté via une matrice 2d à valeurs abritraires (soit None soit Nombre entier)
    représentant salles et couloirs

    Elle permet de récupérer une matrice (objet type IntegerMatrix) constante et adéquate
    pour le but recherché (disposer dun labyrinthe différent à chaque fois)
    via la méthode RandomMaze.getMatrix
    """
    def __init__(self, w, h, min_room_size, max_room_size, density_factor=140):
        self.int_matrix = struct.IntegerMatrix((w, h))
        self.int_matrix.set_all(None)  # super important! Otherwise the algorithm wont work

        self.room_possib_size = list()

        for i in range(min_room_size, max_room_size + 1):
            if i % 2 == 0:  # choix dune taille impaire nécessairement
                continue
            self.room_possib_size.append(i)

        # --- pr stocker meta données
        self.curr_region = 0
        self.li_rooms = list()

        # --- procédure de constr dun labyrinthe randomize

        # (1) creation non overlaping rooms
        self.nb_rooms = 0
        self.all_room_codes = set()
        for i in range(density_factor):
            self.__add_room()

        # (2) growing tree algo
        for i in range(1, w - 1, 2):
            for j in range(1, h - 1, 2):
                pos = (i, j)
                if self.int_matrix.get_val(*pos) is None:
                    self.__growMaze(pos)

        # (3) connexion de regions (creuse les jonctions)
        self.candidats_connexion = list()
        self.li_connecteurs = list()
        dim = self.int_matrix.get_size()
        for i in range(dim[0]):
            for j in range(dim[1]):
                if self.__canBeConnector((i, j)):
                    self.candidats_connexion.append((i, j))

        self.regions_to_blobs = dict()
        while self.canMerge():
            self.stepMerge()

        # (4) suppr. cul-de-sac
        self.tested_pos = set()
        self.recUncarve((1, 1))

    def pick_walkable_cell(self):
        # helps choosing a valid initial position for the avatar
        floor = self.int_matrix
        w, h = floor.get_size()
        chosen_pos = [
            random.randint(0, w-1),
            random.randint(0, h-1)
        ]
        while floor.get_val(*chosen_pos) is None:  # i.e. a wall
            # try again
            chosen_pos[0] = random.randint(0, w-1)
            chosen_pos[1] = random.randint(0, h-1)
        return chosen_pos
    
    def isRoomRegion(self, pos):
        val = self.int_matrix.get_val(*pos)
        return val in self.all_room_codes

    def recUncarve(self, pos):
        if pos in self.tested_pos:
            return
        self.tested_pos.add(pos)

        m_size = self.int_matrix.get_size()
        voisins = cell_voisinnes(pos, m_size)
        nb_cell_carved_v = 0  # sera un nb entre 1 et 4
        for v in voisins:
            if self.int_matrix.get_val(*v) is not None:
                nb_cell_carved_v += 1

        if nb_cell_carved_v == 1:
            self.int_matrix.set_val(pos[0], pos[1], None)
            for v in voisins:
                if v in self.tested_pos:
                    self.tested_pos.remove(v)

        for v in voisins:
            if self.int_matrix.get_val(*v) is not None:
                self.recUncarve(v)

    def __detRegionsProches(self, c):
        # -- traitement basique : lister les codes distincts des régions voisines
        m_size = self.int_matrix.get_size()
        voisins = cell_voisinnes(c, m_size)
        ens_tmp = set()
        for v in voisins:
            tmp_val = self.int_matrix.get_val(*v)
            if tmp_val is not None:
                ens_tmp.add(tmp_val)
        return list(ens_tmp)

    def __canBeConnector(self, pos):
        # un connecteur : est avant tt un mur, il doit letre
        tmp_val = self.int_matrix.get_val(*pos)
        if tmp_val is not None:
            return False

        # un connecteur : est proche de 2 régions distinctes
        codes_reg_prox = self.__detRegionsProches(pos)
        return len(codes_reg_prox) >= 2

    def __isCloseToRegion(self, pos, code):
        codes_reg_prox = self.__detRegionsProches(pos)
        return code in codes_reg_prox

    def canMerge(self):
        return len(self.candidats_connexion) > 0

    def stepMerge(self):
        COEFF_REDONDANCE = 0.3

        connect_valid = random.choice(self.candidats_connexion)
        tmp = self.__detRegionsProches(connect_valid)

        if tmp[0] in self.all_room_codes:  # on privilegie un connecteur = region de couloir
            args = [tmp[1], tmp[0]]
        else:
            args = tmp

        self.__mergeRegions(connect_valid, args[0], args[1])

        # --- On garde que les connecteurs utiles et pas trop près dun connecteur validé
        liste_filtree = list()

        for c in self.candidats_connexion:
            # -- élimination des connecteurs trop proches du nouveau connect_valid
            if dist_manhattan(c, connect_valid) <= 1:
                continue

            # -- utile pr merger?
            merging_utility = False
            cod_rp = self.__detRegionsProches(c)
            if (cod_rp[0] not in self.regions_to_blobs) or (cod_rp[1] not in self.regions_to_blobs):
                merging_utility = True
            elif self.regions_to_blobs[cod_rp[0]] != self.regions_to_blobs[cod_rp[1]]:
                merging_utility = True

            if merging_utility:
                liste_filtree.append(c)
                continue
            # -- conservation "chanceuse" de candidats inutiles
            if random.random() < COEFF_REDONDANCE:
                liste_filtree.append(c)

        self.candidats_connexion = liste_filtree

    def __findCloseWalls(self, i, j):
        murs_voisins = list()
        for d in DIRS:
            tmp = (i + COORD_OFFSET[d][0], j + COORD_OFFSET[d][1])
            if self.int_matrix.get_val(*tmp) is None:
                murs_voisins.append(tmp)
        return murs_voisins

    def __mergeRegions(self, current_cell, cod_region_a, cod_region_b):
        self.int_matrix.set_val(current_cell[0], current_cell[1], cod_region_a)

        self.candidats_connexion.remove(current_cell)
        self.li_connecteurs.append(current_cell)

        candidats_blob = list()
        if cod_region_a in self.regions_to_blobs:
            candidats_blob.append(self.regions_to_blobs[cod_region_a])
        if cod_region_b in self.regions_to_blobs:
            candidats_blob.append(self.regions_to_blobs[cod_region_b])

        if len(candidats_blob) == 0:
            # nouveau blob
            tmp_blob_codes = list(self.regions_to_blobs.values())
            if len(tmp_blob_codes) == 0:
                blob_corresp = 1
            else:
                blob_corresp = max(tmp_blob_codes) + 1
        else:
            # blob existant, on garde le num le plus petit
            blob_corresp = min(candidats_blob)

        self.regions_to_blobs[cod_region_a] = blob_corresp
        self.regions_to_blobs[cod_region_b] = blob_corresp

    def getMatrix(self):
        return self.int_matrix

    def __startRegion(self):
        self.curr_region += 1

    def getRegion(self):
        return self.curr_region

    def __add_room(self):
        taille_room = random.choice(self.room_possib_size)
        w, h = self.int_matrix.get_size()

        bsupw = w - taille_room
        bsuph = h - taille_room

        intervalle_x = [2 * x + 1 for x in range(0, bsupw // 2)]
        intervalle_y = [2 * y + 1 for y in range(0, bsuph // 2)]
        intervalle_x.pop()
        intervalle_y.pop()

        pos_salle = (
            random.choice(intervalle_x),
            random.choice(intervalle_y)
        )

        # salle empiete sur qq chose ?
        for i in range(pos_salle[0] - 1, pos_salle[0] + taille_room + 2):
            for j in range(pos_salle[1] - 1, pos_salle[1] + taille_room + 2):
                if self.int_matrix.get_val(i, j) is not None:
                    return

        # --- création room
        self.__startRegion()
        code = self.getRegion()
        self.li_rooms.append((pos_salle, taille_room, code))

        for i in range(pos_salle[0], pos_salle[0] + taille_room):
            for j in range(pos_salle[1], pos_salle[1] + taille_room):
                self.int_matrix.set_val(i, j, code)
        self.nb_rooms += 1
        self.all_room_codes.add(code)

    def __carve(self, pos):
        self.int_matrix.set_val(pos[0], pos[1], self.getRegion())

    def __canCarve(self, from_pos, direction):
        base_offset = COORD_OFFSET[direction]
        # on dépasse de la matr?
        offset = map(lambda x: x * 3, base_offset)
        offset = tuple(offset)
        dest = (from_pos[0] + offset[0], from_pos[1] + offset[1])

        m_size = self.int_matrix.get_size()
        if not cell_in_range(dest, m_size):
            return False

        # on tente de carve de deja carvé?
        offset = map(lambda x: x * 2, base_offset)
        offset = tuple(offset)
        dest = (from_pos[0] + offset[0], from_pos[1] + offset[1])
        return self.int_matrix.get_val(*dest) is None

    def __growMaze(self, pos):
        last_dir = None
        self.__startRegion()
        self.__carve(pos)
        total_cases_couloir = 1
        self.li_cells_to_explore = [pos]

        while len(self.li_cells_to_explore) > 0:

            index_last = len(self.li_cells_to_explore)
            cell = self.li_cells_to_explore[index_last - 1]

            unmade_possib = list()

            for d in DIRS:
                if self.__canCarve(cell, d):
                    unmade_possib.append(d)

            if len(unmade_possib) > 0:
                if last_dir is None:
                    direction = random.choice(unmade_possib)
                else:
                    if (last_dir not in unmade_possib) or (random.random() < WINDING_FACTOR):
                        direction = random.choice(unmade_possib)
                    else:
                        direction = last_dir

                base_offset = COORD_OFFSET[direction]
                self.__carve((cell[0] + base_offset[0], cell[1] + base_offset[1]))
                cell_creusee = (cell[0] + 2 * base_offset[0], cell[1] + 2 * base_offset[1])
                self.__carve(cell_creusee)
                total_cases_couloir += 2
                if total_cases_couloir <= SEUIL_CASES:
                    self.li_cells_to_explore.append(cell_creusee)
                    last_dir = direction
            else:
                del self.li_cells_to_explore[index_last - 1]
                last_dir = None  # this path ended


# ----------------------------
#   BELOW is an extension by Travis Moy'
#
# *MIT License*
# Copyright (c) 2017 Travis Moy'

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Holds the three angles for each cell. Near is closest to the horizontal/vertical line, and far is furthest.
# Also used for obstructions; for the purposes of obstructions, the center variable is ignored.
class CellAngles(object):
    def __init__(self, near, center, far):
        self.near = near
        self.center = center
        self.far = far

    def __repr__(self):
        return "(near={0} center={1} far={2})".format(self.near, self.center, self.far)


class FOVCalc(object):

    # Changing the radius-fudge changes how smooth the edges of the vision bubble are.
    #
    # RADIUS_FUDGE should always be a value between 0 and 1.
    RADIUS_FUDGE = 1.0 / 3.0

    # If this is False, some cells will unexpectedly be visible.
    #
    # For example, let's say you you have obstructions blocking (0.0 - 0.25) and (.33 - 1.0).
    # A far off cell with (near=0.25, center=0.3125, far=0.375) will have both its near and center unblocked.
    #
    # On certain restrictiveness settings this will mean that it will be visible, but the blocks in front of it will
    # not, which is unexpected and probably not desired.
    #
    # Setting it to True, however, makes the algorithm more restrictive.
    NOT_VISIBLE_BLOCKS_VISION = True

    # Determines how restrictive the algorithm is.
    #
    # 0 - if you have a line to the near, center, or far, it will return as visible
    # 1 - if you have a line to the center and at least one other corner it will return as visible
    # 2 - if you have a line to all the near, center, and far, it will return as visible
    #
    # If any other value is given, it will treat it as a 2.
    RESTRICTIVENESS = 1

    # If VISIBLE_ON_EQUAL is False, an obstruction will obstruct its endpoints. If True, it will not.
    #
    # For example, if there is an obstruction (0.0 - 0.25) and a square at (0.25 - 0.5), the square's near angle will
    # be unobstructed in True, and obstructed on False.
    #
    # Setting this to False will make the algorithm more restrictive.
    VISIBLE_ON_EQUAL = True

    # Parameter func_transparent is a function with the sig: boolean func(x, y)
    # It should return True if the cell is transparent, and False otherwise.
    #
    # Returns a set with all (x, y) tuples visible from the centerpoint.
    def calc_visible_cells_from(self, x_center, y_center, radius, func_transparent):
        cells = self._visible_cells_in_quadrant_from(x_center, y_center, 1, 1, radius, func_transparent)
        cells.update(self._visible_cells_in_quadrant_from(x_center, y_center, 1, -1, radius, func_transparent))
        cells.update(self._visible_cells_in_quadrant_from(x_center, y_center, -1, -1, radius, func_transparent))
        cells.update(self._visible_cells_in_quadrant_from(x_center, y_center, -1, 1, radius, func_transparent))
        cells.add((x_center, y_center))
        return cells

    # Parameters quad_x, quad_y should only be 1 or -1. The combination of the two determines the quadrant.
    # Returns a set of (x, y) tuples.
    def _visible_cells_in_quadrant_from(self, x_center, y_center, quad_x, quad_y, radius, func_transparent):
        cells = self._visible_cells_in_octant_from(x_center, y_center, quad_x, quad_y, radius, func_transparent, True)
        cells.update(self._visible_cells_in_octant_from(x_center, y_center, quad_x, quad_y, radius, func_transparent,
                                                        False))
        return cells

    # Returns a set of (x, y) typles.
    # Utilizes the NOT_VISIBLE_BLOCKS_VISION variable.
    def _visible_cells_in_octant_from(self, x_center, y_center, quad_x, quad_y, radius, func_transparent, is_vertical):
        iteration = 1
        visible_cells = set()
        obstructions = list()

        # End conditions:
        #   iteration > radius
        #   Full obstruction coverage (indicated by one object in the obstruction list covering the full angle from 0
        #      to 1)
        while iteration <= radius and not (len(obstructions) == 1 and
                                           obstructions[0].near == 0.0 and obstructions[0].far == 1.0):
            num_cells_in_row = iteration + 1
            angle_allocation = 1.0 / float(num_cells_in_row)

            # Start at the center (vertical or horizontal line) and step outwards
            for step in range(iteration + 1):
                cell = self._cell_at(x_center, y_center, quad_x, quad_y, step, iteration, is_vertical)

                if self._cell_in_radius(x_center, y_center, cell, radius):
                    cell_angles = CellAngles(near=(float(step) * angle_allocation),
                                             center=(float(step + .5) * angle_allocation),
                                             far=(float(step + 1) * angle_allocation))

                    if self._cell_is_visible(cell_angles, obstructions):
                        visible_cells.add(cell)
                        if not func_transparent(cell[0], cell[1]):
                            obstructions = self._add_obstruction(obstructions, cell_angles)
                    elif self.NOT_VISIBLE_BLOCKS_VISION:
                        obstructions = self._add_obstruction(obstructions, cell_angles)

            iteration += 1

        return visible_cells

    # Returns a (x, y) tuple.
    def _cell_at(self, x_center, y_center, quad_x, quad_y, step, iteration, is_vertical):
        if is_vertical:
            cell = (x_center + step * quad_x, y_center + iteration * quad_y)
        else:
            cell = (x_center + iteration * quad_x, y_center + step * quad_y)
        return cell

    # Returns True/False.
    def _cell_in_radius(self, x_center, y_center, cell, radius):
        cell_distance = math.sqrt((x_center - cell[0]) * (x_center - cell[0]) +
                                  (y_center - cell[1]) * (y_center - cell[1]))
        return cell_distance <= float(radius) + self.RADIUS_FUDGE

    # Returns True/False.
    # Utilizes the VISIBLE_ON_EQUAL and RESTRICTIVENESS variables.
    def _cell_is_visible(self, cell_angles, obstructions):
        near_visible = True
        center_visible = True
        far_visible = True

        for obstruction in obstructions:
            if self.VISIBLE_ON_EQUAL:
                if obstruction.near < cell_angles.near < obstruction.far:
                    near_visible = False
                if obstruction.near < cell_angles.center < obstruction.far:
                    center_visible = False
                if obstruction.near < cell_angles.far < obstruction.far:
                    far_visible = False
            else:
                if obstruction.near <= cell_angles.near <= obstruction.far:
                    near_visible = False
                if obstruction.near <= cell_angles.center <= obstruction.far:
                    center_visible = False
                if obstruction.near <= cell_angles.far <= obstruction.far:
                    far_visible = False

        if self.RESTRICTIVENESS == 0:
            return center_visible or near_visible or far_visible
        elif self.RESTRICTIVENESS == 1:
            return (center_visible and near_visible) or (center_visible and far_visible)
        else:
            return center_visible and near_visible and far_visible

    # Generates a new list by combining all old obstructions with the new one (removing them if they are combined) and
    # adding the resulting obstruction to the list.
    #
    # Returns the generated list.
    def _add_obstruction(self, obstructions, new_obstruction):
        new_object = CellAngles(new_obstruction.near, new_obstruction.center, new_obstruction.far)
        new_list = [o for o in obstructions if not self._combine_obstructions(o, new_object)]
        new_list.append(new_object)
        return new_list

    # Returns True if you combine, False otherwise
    def _combine_obstructions(self, old, new):
        # Pseudo-sort; if their near values are equal, they overlap
        if old.near < new.near:
            low = old
            high = new
        elif new.near < old.near:
            low = new
            high = old
        else:
            new.far = max(old.far, new.far)
            return True

        # If they overlap, combine and return True
        if low.far >= high.near:
            new.near = min(low.near, high.near)
            new.far = max(low.far, high.far)
            return True
        return False



from ... import _hub

pygame = _hub.pygame  # alias to keep on using pygame, easily

import math


def get_line(x1, y1, x2, y2):
    # Bresenham's line drawing algorithm, as obtained from RogueBasin.
    points = []
    issteep = abs(y2 - y1) > abs(x2 - x1)
    if issteep:
        x1, y1 = y1, x1
        x2, y2 = y2, x2
    rev = False
    if x1 > x2:
        x1, x2 = x2, x1
        y1, y2 = y2, y1
        rev = True
    deltax = x2 - x1
    deltay = abs(y2 - y1)
    error = int(deltax / 2)
    y = y1
    ystep = None
    if y1 < y2:
        ystep = 1
    else:
        ystep = -1
    for x in range(x1, x2 + 1):
        if issteep:
            points.append((y, x))
        else:
            points.append((x, y))
        error -= deltay
        if error < 0:
            y += ystep
            error += deltax
    # Reverse the list if the coordinates were reversed
    if rev:
        points.reverse()
    return points


def get_fline(p1, p2, speed):
    # Generate a line, but of floats, ending with the ints x2,y2.
    points = list()
    rng = math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)
    steps = int(rng / speed)
    fsteps = float(rng / speed)
    for t in range(0, steps):
        newpoint = list()
        for coord in range(len(p1)):
            newpoint.append(p1[coord] + float((p2[coord] - p1[coord]) * t) / fsteps)
        points.append(newpoint)
    points.append(p2)
    return points


class MoveModel(object):
    def __init__(self, model, dest, speed=0.1, delay=0):
        self.model = model
        self.speed = speed
        self.dest = dest
        self.delay = delay
        self.step = 0
        self.needs_deletion = False
        self.children = list()
        start = (model.x, model.y)
        self.itinerary = get_fline(start, dest, speed)

    def update(self):
        # This one doesn't appear directly, but moves a model.
        if self.delay > 0:
            self.delay += -1
        elif self.itinerary:
            self.model.x, self.model.y = self.itinerary.pop(0)
            if not self.itinerary:
                self.needs_deletion = True
        else:
            self.needs_deletion = True


class WatchMeWiggle(object):
    # Bear with me. When a model attacks, sometimes it's not clear which model is attacking. So,
    # what I'm gonna do is wiggle the model doing the action, so you can see who it is.
    # Tried to think of a clever name but "WatchMeWiggle" was the best I came up with. You see, the model
    # wiggles, because that's the model you're supposed to be watching right now.
    def __init__(self, model, delay=0, duration=5):
        self.model = model
        self.duration = duration
        self.delay = delay
        self.step = 0
        self.needs_deletion = False
        self.children = list()

    WIGGLE_POS = ((0, 1), (0, 2), (0, 1), (0, 0), (0, -1), (0, -2), (0, -1), (0, 0))

    def update(self, view):
        # This one doesn't appear directly, but moves a model.
        if self.delay > 0:
            self.delay += -1
        elif self.duration > self.step:
            self.step += 1
            self.model.offset_pos = self.WIGGLE_POS[self.step % len(self.WIGGLE_POS)]
        else:
            self.model.offset_pos = None
            self.needs_deletion = True


class BlastOffAnim(object):
    # The model will fly up, up, up for around 1000 pixels. It does not come down again so if you want that you better
    # do it manually.
    def __init__(self, model, delay=0, duration=50, speed=-1, acceleration=-1):
        self.model = model
        self.duration = duration
        self.height = 0
        self.speed = speed
        self.acceleration = acceleration
        self.delay = delay
        self.step = 0
        self.needs_deletion = False
        self.children = list()

    def update(self, view):
        # This one doesn't appear directly, but moves a model.
        if self.delay > 0:
            self.delay += -1
        elif self.duration > self.step:
            self.step += 1
            self.height += self.speed
            self.speed += self.acceleration
            self.model.offset_pos = (0, self.height)
        else:
            self.needs_deletion = True



from ..Singleton import Singleton
from ..util import drawtext


@Singleton
class GameModeMger:
    def __init__(self):
        self.menus = dict()
        self._mode = self.menu = None

    @property
    def mode(self):
        return self._mode

    def set_curr_mode(self, mode_name):
        self._mode = mode_name
        self.menu = self.menus[mode_name]

    def register(self, dictobj):
        self.menus.update(dictobj)

    def switch_mode(self, mode, reset=False):
        if mode in self.menus:
            self._mode = mode
            self.menu = self.menus[self.mode]
            if reset:
                self.menu.__init__(self.menu.name)

    def update(self, events):
        self.menu.update(events)

    def draw(self, surf):
        self.menu.draw(surf)


class BaseGameMode:
    """
    Base signature for all menus
    """

    def __init__(self, name='menu'):
        self._manager = GameModeMger.instance()
        self.name = name

    def update(self, events):  # list[pygame.event.Event]):
        pass

    def draw(self, surf):
        surf.blit(drawtext(self.name, aliased=True), (50, 50))




from setuptools import setup, find_packages
from pkgutil import walk_packages
import katagames_sdk


setup(
    name='katasdk',
    packages = find_packages(include=[
        'katagames_sdk',
        'katagames_sdk.katagames_engine',
        'katagames_sdk.katagames_engine.caps',

        'katagames_sdk.katagames_engine.caps.bioslike',
        'katagames_sdk.katagames_engine.caps.engine_ground',
        'katagames_sdk.katagames_engine.caps.networking',
        'katagames_sdk.katagames_engine.caps.struct',

        'katagames_sdk.katagames_engine.compo',

        'katagames_sdk.katagames_engine.foundation',
        'katagames_sdk.katagames_engine.ifaces',
        'katagames_sdk.katagames_engine.looparts',

        'katagames_sdk.katagames_engine.looparts.ai',
        'katagames_sdk.katagames_engine.looparts.demolib',
        'katagames_sdk.katagames_engine.looparts.gui',
        'katagames_sdk.katagames_engine.looparts.isometric',
        'katagames_sdk.katagames_engine.looparts.polarbear',
        'katagames_sdk.katagames_engine.looparts.tmx',
        ]),
    #entry_points ={
    #    'console_scripts': [
    #    'katasdk = katagames_sdk.katasdk_cmd_line:main'
    #    ]
    #},
    version=katagames_sdk.SDKVER_TAG,  # -- version fournie ici
    include_package_data=True,  # see MANIFEST.in
    description='SDK aimed at creating games for the Kata.games platform',
    author='Gaudia Tech Inc.',
    url='https://github.com/gaudiatech/katagames-glob-repo',
    author_email='thomas@gaudia-tech.com',
    long_description='SDK aimed at creating games for the Kata.games platform, you can find more info on github: user **gaudiatech**',
    license='MIT',
    tests_require=['pytest==4.4.1'],
    install_requires=[
        'pygame>=2.1.2',
        'requests>=2.27.0'
    ],
    test_suite='tests',
)



from ... import _hub as inj
from ... import event


pygame = inj.pygame
EventReceiver = event.EventReceiver


class ButtonPanel(EventReceiver):

    def __init__(self, button_list=None, callbacks=None):
        super().__init__()
        self._lb = list()
        if button_list:
            self._lb.extend(button_list)

        if callbacks:
            self._callbacks = callbacks
        else:
            self._callbacks = dict()

    def proc_event(self, ev, source):
        if ev.type == pygame.MOUSEBUTTONDOWN:
            for bobj in self._lb:
                if bobj.rect.collidepoint(ev.pos):
                    try:
                        self._callbacks[bobj.ident]()
                    except KeyError:
                        self.pev(kataen.EngineEvTypes.BTCLICK, bt_ident=bobj.ident)


class Button(pygame.sprite.Sprite):

    free_bt_identifier = 77348

    def __init__(self, pos=None, size=None, label='bt_label'):
        super().__init__()
        self.callback = None

        self.ident = Button.free_bt_identifier
        Button.free_bt_identifier += 1

        if size:
            self._w, self._h = size
        if pos:
            self.pos = list(pos)
        self.image = pygame.surface.Surface((self._w, self._h))
        self.image.fill('gray')

        tmp_font = pygame.font.Font(None, 18)
        self._lbl = tmp_font.render(label, True, 'navyblue')
        self.label = label
        offsetx = self._lbl.get_width() // 2
        offsety = self._lbl.get_height() // 2
        cx, cy = self._w//2, self._h//2
        self.image.blit(self._lbl, (cx - offsetx, cy - offsety))

        pygame.draw.rect(self.image, 'navyblue', (0, 0, self._w-1, self._h-1), 2)

        # adjust position
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos



from ... import _hub
from ...compo import vscreen as core


ANCHOR_UPPERLEFT = (0, 0)
ANCHOR_UPPERRIGHT = (2, 0)
ANCHOR_CENTER = (1, 1)
ANCHOR_LOWERLEFT = (0, 2)
ANCHOR_LOWERRIGHT = (2, 2)
ANCHOR_TOP = (1, 0)
ANCHOR_LEFT = (0, 1)
ANCHOR_RIGHT = (2, 1)

pygame = _hub.pygame


class Frect(object):
    """Floating rect- changes position depending on the screen dimensions."""

    def __init__(self, dx, dy, w, h, anchor=ANCHOR_CENTER, parent=None):
        self.dx = dx
        self.dy = dy
        self.w = w
        self.h = h
        self.anchor = anchor
        self.parent = parent

    def get_rect(self):
        if self.parent:
            prect = self.parent.get_rect()
            x0 = prect.left + (prect.w // 2) * self.anchor[0]
            y0 = prect.top + (prect.h // 2) * self.anchor[1]
        else:
            sw, sh = core.get_screen().get_size()
            x0 = (sw // 2) * self.anchor[0]
            y0 = (sh // 2) * self.anchor[1]
        return pygame.Rect(self.dx + x0, self.dy + y0, self.w, self.h)



import abc


# --- trick from stackoverflow...
class abstractclassmethod(classmethod):
    __isabstractmethod__ = True

    def __init__(self, fcallable):
        callable.__isabstractmethod__ = True
        super(abstractclassmethod, self).__init__(fcallable)


class BaseMatrix(object):
    """
    classe ABSTRAITE
    représentation 1d dune matrice bidimensionnelle de taille connue
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, matr_size, li1d_val_init=None):
        self.width, self.height = matr_size
        nb_elem = self.width * self.height
        if li1d_val_init is not None:
            if len(li1d_val_init) != nb_elem:
                raise ValueError('val pour init matrice incoherentes')
            for elt in li1d_val_init:
                if not self.__class__.is_valid_val(elt):
                    raise ValueError('val type doesnt match matrix type, elt={}'.format(elt))
            self.repr_1d = li1d_val_init
        else:
            self.repr_1d = [self.__class__.default_val()] * nb_elem

    @abstractclassmethod
    def default_val(cls):
        pass

    @abstractclassmethod
    def is_valid_val(cls, val):
        pass

    def get_val(self, i, j):
        if self.is_out(i, j):
            raise ValueError('coords {} {} out of bounds for the current matrix'.format(i, j))
        adhoc_ind = j * self.width + i
        return self.repr_1d[adhoc_ind]

    def set_val(self, i, j, val):
        if not self.__class__.is_valid_val(val):
            raise ValueError(
                'val {} incompatible avec le type de matrice utilisé ({})'.format(val, self.__class__.__name__)
            )
        ind = j * self.width + i
        self.repr_1d[ind] = val

    def set_all_values(self, val):
        for i in range(self.width):
            for j in range(self.height):
                self.set_val(i, j, val)

    def is_out(self, i, j):
        return not (0 <= i < self.width and 0 <= j < self.height)

    def get_size(self):
        return self.width, self.height

    def __str__(self):
        res = '{} x {} matrix'.format(self.width, self.height)
        res += "\n"
        for j in range(self.height):
            for i in range(self.width):
                ind = j * self.width + i
                res += '  {}'.format(self.repr_1d[ind])
            res += "\n"
        return res


class IntegerMatrix(BaseMatrix):
    """
    modélise une matrice d'entiers. Par défaut, toutes les cellules sont à 0
    """

    # --- redefinition de 2 methodes venant den haut
    @classmethod
    def default_val(cls):
        return None  # TODO faire en sorte que ce soit 0 mais faudra modifer RandomMaze avant cela

    @classmethod
    def is_valid_value(cls, val):
        return (val is None) or isinstance(val, int)



import urllib.parse as urlparse
import urllib.request as urll
from urllib.parse import urlencode
from ...engine.foundation import conf_eng as cgmconf
from ...alpha_pyg.Singleton import Singleton
from ...engine import runs_in_web


@Singleton
class HttpServer:
    def __init__(self):
        self._dev_mode = True
        self._host_str = None
    #     self._default_php_script = None
    #
    # def set_default_script(self, url):
    #     self._default_php_script = url

    def get_gtm_app_url(self):
        return 'https://games.gaudia-tech.com/gtm_app/'

    def get_ludo_app_url(self):
        return 'https://games.gaudia-tech.com/ludo_app/'

    def set_debug_info(self, val):
        self._dev_mode = val

    def __send_get_request(self, url):
        if self._dev_mode:
            print()
            print('---DEBUG NETW.---')
            print('send_get_request')
            print('method GET proxied')
            print(str(url))
            print('...')

        if cgmconf.runs_in_web():
            from browser import ajax, window

            temppp = '0x999'
            def ma_fonc(result):
                nonlocal temppp  # access the outer scope
                # window.console.log('kikoooooo on a vrai resulta la' + str(result))
                temppp = result.text

            ajax.get(url, blocking=True, oncomplete=ma_fonc)
            res = temppp  # json
        else:
            req = urll.Request(url)
            response = urll.urlopen(req)
            res = response.read().decode('ascii')

        if self._dev_mode:
            print(res)
            print('-----------------')
        return res

    def proxied_post(self, url_script_php, post_fields: dict):
        if self._dev_mode:
            print()
            print('---DEBUG NETW.---')
            print('method POST proxied yyaaaa')
            print(str(url_script_php))
            print(str(post_fields))
            print('...')

        if runs_in_web():
            from browser import ajax, window

            temppp = '0x999'
            def ma_fonc(result):
                nonlocal temppp  # access the outer scope
                # window.console.log('kikoooooo on a vrai resulta la' + str(result))
                # window.console.log(str(result.status))
                temppp = result.text

            window.console.log(post_fields)
            ajax.post(url_script_php, data=post_fields, blocking=True, oncomplete=ma_fonc)
            res = temppp  # json

        else:
            # params = pp.urlencode(post_data_dict)
            # headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
            request = urll.Request(url_script_php, urlencode(post_fields).encode())
            res = urll.urlopen(request).read().decode('ascii')

        if self._dev_mode:
            print(res)
            print('-----------------')

        return res  # -> json format str

    # def remote_func(self, rem_func_name, dico_params, url_script_php=None):
    #     """
    #     fonction générique envoyant une requête HTTP vers le serveur,
    #     indispensable, mais quasi-jamais utilisée ! (usage déconseillé)
    #
    #     :param url_script_php: URL + script PHP à exécuter e.g. http://games.gaudia-tech.com/brutos/legacy/bopvp.php
    #     :param rem_func_name: valeur pr le paramètre spécial "func", détermine le cas à traiter côté serveur
    #     :param dico_params: repr. des associations nom_param_get <> valeur_pour_ce_param
    #     :return: réponse HTTP brute, peut-être None
    #     """
    #
    #     if url_script_php is None:
    #         assert self._default_php_script
    #         url_script_php = self._default_php_script
    #
    #     if dico_params is None:
    #         dico_params = dict()
    #
    #     if rem_func_name is not None:
    #         dico_params['func'] = rem_func_name
    #
    #     if len(dico_params) > 0:
    #         suffix = urlparse.urlencode(dico_params)
    #         a_envoyer = url_script_php + '?' + suffix  # signe de la methode GET
    #     else:
    #         a_envoyer = url_script_php
    #
    #     return self.__send_get_request(a_envoyer)

    def proxied_get(self, url, get_params):
        if len(get_params) > 0:
            suffix = urlparse.urlencode(get_params)
            a_envoyer = url + '?' + suffix  # signe de la methode GET
        else:
            a_envoyer = url
        if self._dev_mode:
            print(a_envoyer)
        return self.__send_get_request(a_envoyer)



import re


def underscore_format(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def camel_case_format(string_ac_underscores):
    words = [word.capitalize() for word in string_ac_underscores.split('_')]
    return "".join(words)



"""
Barebones isometric_map handling for an isometric RPG. For game-specific data,
either subclass the Scene or just declare whatever extra bits are needed.
"""

from .IsometricMapViewer import IsometricMapViewer
from .IsometricMapViewer0 import IsometricMapViewer0
from . import extras
from . import model


def set_tiled_version(vernum):
    if float(vernum) < 1.9:
        model.info_type_obj = 'type'
        print('*LEGACY isometric model ON*')



import random

from .FSA_classes_base import Automate, Etat, Transition

MSG_NEUTRES = [
    "Qu'est-ce que vous m'racontez bon sang ?",
    "J'en sais fichtre rien...",
    "J'vois pas de quoi vous causez, là."
]
MSG_BLOCAGE_NEG = [
    "...",
    "Les aigles ne volent pas avec les pigeons.",
    "Peu importe. Z'êtes pas quelqu'un dont les dires m'intéressent.",
    "Voilà ce qu'on va faire : vous allez là-bas, moi j'reste ici.",
    "Adieu."
]
MSG_BLOCAGE_POS = [
    "On parle, on parle, concentrons-nous plutôt sur l'action !",
    "Suis encor' en train de préparer les armes pour notre prochaine quête...",
    "Ah vous revoilà l'ami !",
    "Hola ! Revoilà le champion !",
    "Suis en train de réunir des pièces pour partir à l'aventure avec vous."
]


class NorrecBrain(Automate):

    def __init__(self):

        self.assoc_etat_msg = {
            'bourru': "Ghhrrn qu'est-ce que c'est... Moi c'est Norrec l'guerrier. J'vous ai déjà vu vous ?",
            'faux_oui': "Ça m'étonnerait ! **non** serait la réponse appropriée, cessez de **mentir**",
            'faux_non': "Ça m'étonnerait ! **oui** serait la réponse appropriée, cessez de **mentir**",
            'suspicieux': "Mentir c'est la pire abomination ! C'est l'attitude reservée aux rats ! Vous êt' d'accord avec moi ?",
            'neutre': "Par ici on aime pas trop les vagabonds surtout ceux qui se promènent en **armure** comme la vôtre.",
            'curieux': "Ah c'est sûr que j'en vois pas souvent des comme ça, **oui** remarquez ce doit être **utile**, **non** ?",
            'demandeur': "Vous dites oui c'est utile ! Comme j'vous comprends. Un bel acier comme ça. J'vous connais pas m'enfin... "
                         + "Vous parlez comme quelqu'un de vrai. "
                         + "Vous pouvez peut-être me **renseigner** sur où acheter une telle armure ?",
            'chaud': "Fantastique ! On dirait bien que je me suis trompé sur vot' compte. "
                     + "Vous êtes bien brave. Si vous avez besoin de recruter un homme fort, ma hache est à vot' service.",
            'froid': "Ah, c'est bien ce que j'pensais : une crapule de plus dans not' ville ! Pouah ! (Norrec crache à vos pieds)"
        }

        # --- définition de 9 états, avec les transitions sortantes respectives
        et1 = Etat('bourru', [
            Transition('oui', 'faux_oui'),
            Transition('non', 'neutre')
        ], True, False)

        et2 = Etat('faux_oui', [
            Transition('mentir', 'suspicieux'),
            Transition('oui', 'neutre')
        ])

        et3 = Etat('suspicieux', [
            Transition('oui', 'neutre'),
            Transition('non', 'froid')
        ])

        et4 = Etat('froid', [], False, True)

        et5 = Etat('neutre', [
            Transition('armure', 'curieux')
        ])

        et6 = Etat('curieux', [
            Transition('non', 'faux_non'),
            Transition('utile', 'demandeur'),
            Transition('oui', 'demandeur')
        ])

        et7 = Etat('faux_non', [
            Transition('mentir', 'suspicieux'),
            Transition('oui', 'demandeur')
        ])

        et8 = Etat('demandeur', [
            Transition('non', 'froid'),
            Transition('oui', 'chaud'),
            Transition('renseigner', 'chaud')
        ])

        et9 = Etat('chaud', [], False, True)

        etats_norrec = [et1, et2, et3, et4, et5, et6, et7, et8, et9]
        super().__init__(etats_norrec)

        self.adhoc_msg = self.assoc_etat_msg[
            self.get_etat_courant().get_label()
        ]

    # redéfinition
    def entree_info(self, info):
        etat_avant = self.get_etat_courant()
        super().entree_info(info)
        nouvel_etat = self.get_etat_courant()

        if nouvel_etat == etat_avant:
            if etat_avant.est_terminal():
                if etat_avant.get_label() == 'froid':
                    self.adhoc_msg = random.choice(MSG_BLOCAGE_NEG)
                else:
                    self.adhoc_msg = random.choice(MSG_BLOCAGE_POS)
                return

            self.adhoc_msg = random.choice(MSG_NEUTRES)
            return

        self.adhoc_msg = self.assoc_etat_msg[nouvel_etat.get_label()]

    def say_something(self):
        print(self.adhoc_msg)
